'''
ploting tools for creating the interactive map
for Open Data Cube 

The some function reference from dea_tools library(https://www.dea.ga.gov.au/)

'''

# Import required packages
import math
import folium
import numpy as np
import geopandas as gpd
import json
import os
import datacube 
import xarray as xr
import matplotlib.pyplot as plt 
import odc.ui


from pyproj import Transformer
from odc.ui import image_aspect
from folium.plugins import Draw, Geocoder, Fullscreen, LocateControl
from datacube.utils.cog import write_cog 

# This function is used to plot rgb image from dataset
def rgb(ds,
        bands=['red', 'green', 'blue'],
        index=None,
        index_dim='time',
        robust=True,
        percentile_stretch=None,
        col_wrap=4,
        size=6,
        aspect=None,
        savefig_path=None,
        savefig_kwargs={},
        **kwargs):
    
    # checks bands in dataset or not?
    ds_bands = list(ds.data_vars)
    if set(bands).issubset(ds_bands) == False:
       
        raise ValueError(
            'The bands of your dataset not match with rgb function'
            ' Please check the default is [red, green, blue]') 
    
    #get name of x,y dimension

    try: y_dim, x_dim = ds.geobox.dimension
    except AttributeError: 
        from datacube.utils import spatial_dims
        y_dim, x_dim = spatial_dims(ds)

    #create aspect size kwarg. it will be passed to imshow
    if 'ax' in kwargs:
        aspect_size_kwarg = {}
    else: 
        if not aspect:
            aspact = image_aspect(ds)
            
        aspect_size_kwarg = {'aspect': aspect, 'size': size}

    # <'index' is not provide value>
    #the index defult is none 
    #thus, plot using defult value and arguments passed via `**kwargs`
    if index is None:
        da = ds[bands].to_array()

        if percentile_stretch:
            value_min, valuse_max = da.compute().quantile(
                percentile_stretch).valuse
            
            kwargs.update({'vmin': value_min, 'vmax': valuse_max})

        #if it has more than 3 dims and index_dim = 1
        # squeeze this dimension out to remove it
        if ((len(index)>2) and ('col' not in kwargs) and
            (len(da[index_dim]) == 1)):
            da = da.squeeze(dim = index_dim)

        #if it has more than 3 dims and index_dim > 1
        elif ((len(ds.dims) > 2) and ('col' not in kwargs) and
              (len(da[index_dim]) > 1)):

            raise Exception(
                f'The input dataset `ds` has more than two dimensions: '
                f'{list(ds.dims.keys())}. Please select a single observation '
                'using e.g. `index=0`, or enable faceted plotting by adding '
                'the arguments e.g. `col="time", col_wrap=4` to the function '
                'call')
        da = da.compute()
        img = da.plot.imshow(robust=robust, 
                             col_wrap=col_wrap,
                             **aspect_size_kwarg,
                             **kwargs)
        
    #<If 'index' provide index value >
    else:
        
        # if 'index' give foat num, raise exception to change to int
        if isinstance(index, float):
            raise Exception(
                f'please provide index value as the integer or list of integer'
            )
        
        # if it provide 'col' too! raise exception
        if 'col' in kwargs:
            raise Exception(
                'Cannot provide both \'index\' and \'col\' '
                "Please remove one of it and try again"
            )
        
        # colect index into list to be computed
        index = index if isinstance(index, list) else [index]

        # select band and observations and convert to data array
        da = ds[bands].isel(**{index_dim:index}).to_array().compute()

        if percentile_stretch:
            value_min, valuse_max = da.compute().quantile(
                percentile_stretch).valuse
            
            kwargs.update({'vmin': value_min, 'vmax': valuse_max})
        
        # if multiple index
        if len(index)>1:
            img = da.plot.imshow(x=x_dim,
                                 y=y_dim,
                                 robust=robust,
                                 col=index_dim,
                                 col_wrap=col_wrap,
                                 **aspect_size_kwarg,
                                 **kwargs)

        # If only one index is supplied, squeeze out index_dim and plot as a single panel
        else:

            img = da.squeeze(dim=index_dim).plot.imshow(
                robust=robust, **aspect_size_kwarg, **kwargs)
            

    # If an export path is provided, save image to file. Individual and
    # faceted plots have a different API (figure vs fig) so we get around this
    # using a try statement:
    if savefig_path:

        print(f'Exporting image to {savefig_path}')

        try:
            img.fig.savefig(savefig_path, **savefig_kwargs)
        except:
            img.figure.savefig(savefig_path, **savefig_kwargs)       


def _degree_to_zoom_level(l1, l2, margin = 0.0):

    #Helper function to set zoom level for 'display_map'

    degree = abs(l1 - l2) * (1 + margin)
    zoom_level_int = 0
    if degree != 0:
        zoom_level_float = math.log(360/degree)/math.log(2)
        zoom_level_int = int(zoom_level_float)
    else:
        zoom_level_int = 18
    return zoom_level_int


def display_map(x, y, crs='EPSG:4326'):
    '''
    Generates a Folium map with a latlon bounded rectangle drawn on it.
    Parameters

    '''
    try:
        assert x is not None
        assert y is not None
    except AssertionError as error:
        print(error)
    
    # Convert each corner coordinates to lat-lon
    x_corner = (x[0], x[1], x[0], x[1])
    y_corner = (y[0], y[0], y[1], y[1])
    transformer = Transformer.from_crs(crs,"EPSG:4326")
    lon, lat = transformer.transform(x_corner, y_corner)

    # Calculate zoom level based on coordinates
    margin=-0.5 
    zoom_bias=0
    lat_zoom_level = _degree_to_zoom_level( min(lat), max(lat), margin=margin ) + zoom_bias
    lon_zoom_level = _degree_to_zoom_level(min(lon),max(lon), margin=margin ) + zoom_bias
    zoom_level = min(lat_zoom_level, lon_zoom_level)

    # Identify centre point for plotting
    center = [np.mean(lat), np.mean(lon)]

    # Create map
    Hybrid_map = folium.Map(
        location=center,
        zoom_start=zoom_level,
        tiles="http://mt1.google.com/vt/lyrs=y&z={z}&x={x}&y={y}",
        attr="Google")

    # Create bounding box coordinates to overlay on map
    line_segments = [(lat[0], lon[0]),
                     (lat[1], lon[1]),
                     (lat[3], lon[3]),
                     (lat[2], lon[2]),
                     (lat[0], lon[0])]

    # Add bounding box as an overlay
    Hybrid_map.add_child(folium.features.PolyLine(locations=line_segments,
                                                  color='orange',
                                                  opacity=0.8))

    # Add clickable lat-lon popup box
    Hybrid_map.add_child(folium.features.LatLngPopup())

    return Hybrid_map

# create map from lat/lon
def show_map(lat=None, lon=None, with_draw_tools=True,zoom=None):
    """Create a map using Folium."""
    
    # Set default latitude and longitude values
    if lat is None:
        lat = 13.726  # Bangkok latitude
    if lon is None:
        lon = 100.514  # Bangkok longitude
    
     # set default zoom level
    if zoom is None:
        zoom = 10
             
    # Create a map using Folium
    m = folium.Map(location=[lat, lon], zoom_start=zoom, tiles = None)
    
    # add geocoder plugin
    geocoder = Geocoder(position='topleft', collapsed=True, show=True)
    geocoder.add_to(m)
    
    # add LocateControl plugin
    Lo = LocateControl()
    Lo.add_to(m)
    
    # add fullscreen plugin
    Fullscreen().add_to(m) 
    
    # add masker location
    folium.Marker([lat, lon],
                  icon=folium.Icon(color="red", icon="flag")).add_to(m)
    
    # add WMTS
    # folium.WmsTileLayer('https://basemap.nationalmap.gov/arcgis/services/USGSTopo/MapServer/WMSServer?', 
    #                     layers='0', name = 'Topography', attr='USGS', format= 'image/png', transparent = True).add_to(m)
    
  
    # add tile layer map
    folium.TileLayer('https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}', name='Google Satellite', attr='Google').add_to(m)
    folium.TileLayer('https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}', name='Terrain', attr='Google').add_to(m)
    folium.TileLayer('https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}', name='Hybrid', attr='Google').add_to(m)
    folium.TileLayer('https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png', name='OpenStreetMap', attr='OpenStreetMap').add_to(m)
    folium.TileLayer('https://basemap.sphere.gistda.or.th/tiles/sphere_hybrid/EPSG3857/{z}/{x}/{y}.jpeg?key=42B90819583344A789DA424BE70CDB61', name='Gistda Hybrid', attr='sphere.gistda').add_to(m)
    folium.TileLayer('https://basemap.sphere.gistda.or.th/tiles/thailand_images/EPSG3857/{z}/{x}/{y}.jpeg?key=test2022', name='Gistda Satellite', attr='sphere.gistda').add_to(m)

    # Add layer control to the map
    folium.LayerControl().add_to(m)

    # Add drawing tools to create the polygon
    if with_draw_tools:
        
        draw = Draw(
            export=True,
            filename='drawn_polygons.geojson',
            draw_options={
                'polygon': {'allowIntersection': True},
                'rectangle': True,
                'circle':  True,
                'marker':  True,
                'polyline':  True,
            },
            edit_options={
                'featureGroup': None,
            },
        )

        draw.add_to(m)

    return m

# the function to display the polygons on the display map
def show_vector(file_paths):
     
    # Create a Folium map centered on the first polygon of the first file
    polygons = gpd.read_file(file_paths[0])
    center_lon, center_lat = polygons['geometry'][0].centroid.coords[0]
    m = folium.Map(location=[center_lat, center_lon], zoom_start=13,  tiles = None)
    
    # Define a list of colors to use for each file
    colors = ['lime','blue', 'red', 'yellow','#40E0D0', 'orange', 'CA33FF','Aqua','Fuchsia',
             'lime','blue', 'red', 'yellow','#40E0D0', 'orange', 'CA33FF','Aqua','Fuchsia']
   
    # Read each GeoJSON file and create a GeoDataFrame
    gdfs = [gpd.read_file(file) for file in file_paths]
    
    # Add each GeoDataFrame as a separate layer with a different color
    for i, gdf in enumerate(gdfs):
        filename = os.path.splitext(os.path.basename(file_paths[i]))[0]
        folium.GeoJson(gdf,
                name=  filename,
                style_function=lambda x, color=colors[i]: {'color': color, 'fillOpacity': 0.2},
                tooltip=gdf.columns.tolist()).add_to(m)

                       
    # Add tile map
    folium.TileLayer('https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}', name='Google Satellite', attr='Google').add_to(m)
    folium.TileLayer('https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}', name='Terrain', attr='Google').add_to(m)
    folium.TileLayer('https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}', name='Hybrid', attr='Google').add_to(m)
    folium.TileLayer('https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png', name='OpenStreetMap', attr='OpenStreetMap').add_to(m)
    folium.TileLayer('https://basemap.sphere.gistda.or.th/tiles/sphere_hybrid/EPSG3857/{z}/{x}/{y}.jpeg?key=42B90819583344A789DA424BE70CDB61', name='Gistda Hybrid', attr='sphere.gistda').add_to(m)
    folium.TileLayer('https://basemap.sphere.gistda.or.th/tiles/thailand_images/EPSG3857/{z}/{x}/{y}.jpeg?key=42B90819583344A789DA424BE70CDB61', name='Gistda Satellite', attr='sphere.gistda').add_to(m)
    
    # Add layer control to the map
    folium.LayerControl().add_to(m)

    # Show the map
    return m

# the function to create new column in geojson file
# to add classtye table for training data
def geojson_add_Newcol(geojson_file_path):
    with open(geojson_file_path, 'r') as f:
        geojson_data = json.load(f)

    # Count the number of rows in the GeoJSON data
    
    num_rows = len(geojson_data['features'])
    print('num of rows: ',num_rows)
  
    # Prompt the user for the new column name and value
    new_column_name = input("Enter the name of the new column: ")
    # new_column_value = int(input("Enter the value for the new column: "))
    
    # Try to get an integer input for the new column value
    while True:
        new_column_value = input("Enter the value for the new column: ")
        try:
            new_column_value = int(new_column_value)
            break
        except ValueError:
            print("Invalid input: Please enter an integer value.")
    
    # Add the new column to each feature in the GeoJSON data
    for feature in geojson_data['features']:
        feature['properties'][new_column_name] = new_column_value

    # Write the updated GeoJSON data back to the file
    with open(geojson_file_path, 'w') as f:
        json.dump(geojson_data, f)

    # Return the number of rows in the GeoJSON data before the new column was added
    return num_rows


#load dataset from DC /plot color image from bands/export to GeoTiff
def get_img(query, filename):
   
    # connect to datacube
    dc = datacube.Datacube()
    
    # load dataset from query
    ds = dc.load(**query)
    print (ds)
    print('-'*50)
    a = list(ds.data_vars)[0]
    
    # preview img dataset
    img =ds[a].plot(col="time", robust=True)
    plt.show(img)
    
    # count num of timestep
    dss=len(ds.time)
    
    print ('-'*50)
    print ('There are:_ ' + str(dss) + '_timesteps')
    
    # get timestep from user
    time_step = int(input("Choose the timestep that you want like: [0,1,2,...] "))
    
    if len(query['measurements']) >= 3:
        #plot image selected
        rgb(ds, bands=query["measurements"][:3], index=time_step) 

    else: 
        pass

    
    # combine RGB bands into a single xarray DataArray
    rgb_array = xr.concat([ds[band].isel(time=time_step) for band in query['measurements']], dim='band')
    
    # export selected image to GeoTiff
    write_cog(geo_im=rgb_array, fname=filename, overwrite=True)
        
    # # export selected image to GeoTiff
    # write_cog(geo_im=ds[query['measurements'][0]].isel(time=time_step), fname=filename, overwrite=True)
     
    
    # Format the time stamp for use as the plot title
    time_string = str(ds.time.isel(time=time_step).values).split('.')[0]  

    # Set the title and axis labels
    ax = plt.gca()
    ax.set_title(f"Timestep {time_string}", fontweight='bold', fontsize=16)
    ax.set_xlabel('Easting (m)', fontweight='bold')
    ax.set_ylabel('Northing (m)', fontweight='bold')

    # Display the plot
    plt.show()
    
    
def find_centroid(file, output):
    
    # read input and erase file
    gdf = gpd.read_file(file)
    
    # perform func
    result = gdf.centroid
    
        #save to output
    result.to_file(output)
    
    



'''
Geoprocessing tools for analysing remote sensing data
using in Open Data Cube 
It's consists of:
- convert_geojson_crs
- convert_shp_crs
- merge
- intersect
- union
- erase
- clip
- convert_geojson_to_shp
- convert_shp_to_geojson
'''

import json
import geojson
import pandas as pd
import geopandas as gpd
from shapely.geometry import shape 


        
# function to convert CRS of GeoJson 
def convert_geojson_CRS(input_path, output_path, crs):
    # Read in the GeoJSON file using geopandas
    gdf = gpd.read_file(input_path)

    # Convert the geometry to the desired CRS
    gdf = gdf.to_crs(crs)

    # Write out the GeoDataFrame to a SHP file with the specified CRS and encoding
    gdf.to_file(output_path, driver='GeoJSON', crs=crs, encoding='utf-8')
    
# function to convert CRS of shp 
def convert_shp_CRS(input_path, output_path, crs):
    # Read in the GeoJSON file using geopandas
    gdf = gpd.read_file(input_path)

    # Convert the geometry to the desired CRS
    gdf = gdf.to_crs(crs)

    # Write out the GeoDataFrame to a SHP file with the specified CRS and encoding
    gdf.to_file(output_path, driver='ESRI Shapefile', crs=crs, encoding='utf-8')

# the merge tool combines datasets that are the same datatype into one   
def merge(files, output_file):
    # Initialize an empty list of features
    features = []

    # Loop through all the input files and add their features to the list
    for file in files:
        with open(file) as f:
            data = json.load(f)
            features += data['features']

    # Create a new GeoJSON object
    new_data = {
        "type": "FeatureCollection",
        "features": features
    }

    # Write the new GeoJSON file
    with open(output_file, 'w') as f:
        geojson.dump(new_data, f)
        
        
# function for intersect
def intersect(file_paths, output_file):
    
    # Read each GeoJSON file and create a GeoDataFrame
    gdfs = [gpd.read_file(fp) for fp in file_paths]
    
    # perform intersect on each pair of geodataframe
    result = gdfs[0]
    for gdf in gdfs[1:]:
        result = result.intersection(gdf)
        
    #save result to output file
    result.to_file(output_file)
    
    
# function for union
def union(file_paths, output_file):
    
    # read each files and create a geodataframe
    gdfs = [gpd.read_file(fp) for fp in file_paths]
    
    # perfrom Union on each pair of Geodataframe
    result = gdfs[0]
    for gdf in gdfs[1:]:
        result = result.union(gdf)
    # save to output file
    result.to_file(output_file)
    
    
# func for Erase/Difference    
def erase(file, erase_file, output_file):
    
    # read input and erase file
    gdf = gpd.read_file(file)
    gdfe = gpd.read_file(erase_file)
    
    # perform func
    result = gdf.overlay(gdfe, how = 'difference')
    
    # savs to output
    result.to_file(output_file)
    

#func for clip
def clip(file, clip_file, output_file):
    
    # read input and erase file
    gdf = gpd.read_file(file)
    gdfc = gpd.read_file(clip_file)
    
    # perform func
    result = gpd.clip(gdf, gdfc)
    
    # savs to output
    result.to_file(output_file)
    
#func for Symmetrical difference
def smc_difference(file, diff_file, output_file):
    
    # read input and erase file
    gdf = gpd.read_file(file)
    gdfd = gpd.read_file(diff_file)
    
    # perform function
    result = gdf.overlay(gdfd, how = 'symmetric_difference')
    
    #save to output
    result.to_file(output_file)
    

# func for dissolve    
def dissolve(file, output, dissolve_by):
    
    # read input and erase file
    gdf = gpd.read_file(file)
    
    # perform func
    result = gdf.dissolve(by=dissolve_by)
       
    #save to output
    result.to_file(output)
    
    
# func for buffer ie. buffer = 0.005 (500 meter)
def buffer(file, output, buffer_num):
    
    # read input and erase file
    gdf = gpd.read_file(file)
       
    # perform func
    result = gdf.buffer(buffer_num)
            
    #save to output
    result.to_file(output)

# function to convert GeoJson to Shp
def convert_geojson_to_shp(input_path, output_path, crs):
    # Read in the GeoJSON file using geopandas
    gdf = gpd.read_file(input_path)

    # Convert the geometry to the desired CRS
    gdf = gdf.to_crs(crs)

    # Write out the GeoDataFrame to a SHP file with the specified CRS and encoding
    gdf.to_file(output_path, driver='ESRI Shapefile', crs=crs, encoding='utf-8')
    
# function to convert shp to GeoJSON
def convert_shp_to_geojson(input_path, output_path, crs):
    # Read in the GeoJSON file using geopandas
    gdf = gpd.read_file(input_path)

    # Convert the geometry to the desired CRS
    gdf = gdf.to_crs(crs)

    # Write out the GeoDataFrame to a SHP file with the specified CRS and encoding
    gdf.to_file(output_path, driver='GeoJSON', crs=crs, encoding='utf-8')

    

    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    


import warnings
import numpy as np

def cal_index(ds,
              index=None,
              collection=None,
              custom_varname=None,
              normalise=True,
              drop=False,
              inplace=False):
    
    # Set ds equal to a copy of itself in order to prevent the function 
    # from editing the input dataset. This can prevent unexpected
    # behaviour though it uses twice as much memory. 
    if not inplace:
        ds = ds.copy(deep=True)
    
    # Capture input band names in order to drop these if drop=True
    if drop:
        bands_to_drop=list(ds.data_vars)
        print(f'Dropping bands {bands_to_drop}')

    # Dictionary containing remote sensing index band recipes
    index_dict = {
                  # Normalised Difference Vegation Index, Rouse 1973
                  'NDVI': lambda ds: (ds.nir - ds.red) /
                                     (ds.nir + ds.red),
        
                # Normalised Difference Vegation Index, Rouse 1973
                  'GNDVI': lambda ds: (ds.nir - ds.green) /
                                     (ds.nir + ds.green),

                  # Leaf Area Index, Boegh 2002
                  'LAI': lambda ds: (3.618 * ((2.5 * (ds.nir - ds.red)) /
                                     (ds.nir + 6 * ds.red -
                                      7.5 * ds.blue + 1)) - 0.118), 
        
                  # Ratio vegetation Index, Boegh 2002
                  'RVI': lambda ds: (ds.nir) /
                                     (ds.red), 
        
                  # Simple Ratio (SR) 
                  'SR': lambda ds: ds.nir/ds.red, 
        
                  # Soil Adjusted Vegetation Index, Huete 1988
                  'SAVI': lambda ds: ((1.5 * (ds.nir - ds.red)) /
                                      (ds.nir + ds.red + 0.5)),
        
                  # Modified Soil Adjusted Vegetation Index, Qi et al. 1994
                  'MSAVI': lambda ds: ((2 * ds.nir + 1 - 
                                      ((2 * ds.nir + 1)**2 - 
                                       8 * (ds.nir - ds.red))**0.5) / 2), 
        
                  # Normalised Difference Water Index, McFeeters 1996
                  'NDWI': lambda ds: (ds.green - ds.nir) /
                                     (ds.green + ds.nir),
        
                  # Water Index, Fisher 2016
                  'WI': lambda ds: (1.7204 + 171 * ds.green + 3 * ds.red -
                                    70 * ds.nir - 45 * ds.swir1 -
                                    71 * ds.swir2),
                  # Difference vegetation index (DVI) 
                  'DVI': lambda ds: (ds.nir - ds.red), 
        
                  # Enhanced Vegetation Index, Huete 2002
                  'EVI': lambda ds: ((2.5 * (ds.nir - ds.red)) /
                                     (ds.nir + 6 * ds.red -7.5 * ds.blue + 1)),
        
                  # Burn Area Index, Martin 1998
                  'BAI': lambda ds: (1.0 / ((0.10 - ds.red) ** 2 +
                                            (0.06 - ds.nir) ** 2)) }
    
    
    

    # If index supplied is not a list, convert to list. This allows us to
    # iterate through either multiple or single indices in the loop below
    indices = index if isinstance(index, list) else [index]
    
    #calculate for each index in the list of indices supplied (indexes)
    for index in indices:

        # Select an index function from the dictionary
        index_func = index_dict.get(str(index))

        # If no index is provided or if no function is returned due to an 
        # invalid option being provided, raise an exception informing user to 
        # choose from the list of valid options
        if index is None:

            raise ValueError(f"No remote sensing `index` was provided. Please "
                              "refer to the function \ndocumentation for a full "
                              "list of valid options for `index` (e.g. 'NDVI')")

        elif (index in ['WI','EVI', 'LAI', 'SAVI', 'MSAVI'] 
              and not normalise):

            warnings.warn(f"\nA coefficient-based index ('{index}') normally "
                           "applied to surface reflectance values in the \n"
                           "0.0-1.0 range was applied to values in the 0-10000 "
                           "range. This can produce unexpected results; \nif "
                           "required, resolve this by setting `normalise=True`")

        elif index_func is None:

            raise ValueError(f"The selected index '{index}' is not one of the "
                              "valid remote sensing index options. \nPlease "
                              "refer to the function documentation for a full "
                              "list of valid options for `index`")

        # Rename bands to a consistent format if depending on what collection
        # is specified in `collection`. This allows the same index calculations
        # to be applied to all collections. If no collection was provided, 
        # raise an exception.
        
        dataset = ['ls8_collections_sr_scene']
        if collection is None:

            raise ValueError("'No `collection` was provided. Please specify "
                             "either 'ls8_collections_sr_scene'"
                             "to ensure the function calculates indices using the "
                             "correct spectral bands")

        elif collection in dataset:

            # Dictionary mapping full data names to simpler 'red' alias names
            bandnames_dict = {
                'nir': 'nir',
                'red': 'red',
                'green': 'green',
                'blue': 'blue',
                'swir_1': 'swir1',
                'swir_2': 'swir2',
                'aerosol_qa	': 'aerosol_qa',
                'radsat_qa': 'radsat_qa',
                'pixel_qa': 'pixel_qa'
            }

            # Rename bands in dataset to use simple names (e.g. 'red')
            bands_to_rename = {
                a: b for a, b in bandnames_dict.items() if a in ds.variables
            }


        # Raise error if no valid collection name is provided:
        else:
            raise ValueError(f"'{collection}' is not a valid option for "
                              "`collection`. Please specify either \n"
                              "'ls8_collections_sr_scene'")

        # Apply index function 
        try:
            # If normalised=True, divide data by 10,000 before applying func
            mult = 10000.0 if normalise else 1.0
            index_array = index_func(ds.rename(bands_to_rename) / mult)
        except AttributeError:
            raise ValueError(f'Please verify that all bands required to '
                             f'compute {index} are present in `ds`. \n'
                             f'These bands may vary depending on the `collection` '
                             f'(e.g. the Landsat `nir` band \n'
                             f'is equivelent to `nir_1` for Sentinel 2)')

        # Add as a new variable in dataset
        output_band_name = custom_varname if custom_varname else index
        ds[output_band_name] = index_array
    
    # Once all indexes are calculated, drop input bands if inplace=False
    if drop and not inplace:
        ds = ds.drop(bands_to_drop)

    # If inplace == True, delete bands in-place instead of using drop
    if drop and inplace:
        for band_to_drop in bands_to_drop:
            del ds[band_to_drop]

    # Return input dataset with added water index variable
    return ds


        
                                    
        


import shapely
import xarray as xr
import datacube
import matplotlib
import pydotplus
import numpy as np
import subprocess as sp
import pandas as pd
import geopandas as gpd
import matplotlib.pyplot as plt
from io import StringIO
from odc.io.cgroups import get_cpu_quota
from sklearn.ensemble import RandomForestClassifier
from sklearn import model_selection
from sklearn.metrics import accuracy_score
from IPython.display import Image
from datacube.utils import geometry
from datacube.utils.cog import write_cog
from dea_tools.classification import collect_training_data, predict_xr
from dea_tools.bandindices import calculate_indices
from dream_river.convertools import vectorize
import warnings
warnings.filterwarnings("ignore")

def custom_loadband(query):

    # Initialise datacube
    dc = datacube.Datacube(app='extract_VI')

    # Load data using query
    dd = dc.load(**query)

    # band index
    dd['NDVI']=(dd.nir - dd.red)/(dd.nir + dd.red)
    dd['GNDVI']=(dd.nir - dd.green)/(dd.nir + dd.green)
    dd['EVI']=((2.5 * (dd.nir - dd.red))/(dd.nir + 6 * dd.red -7.5 * dd.blue + 1))
    dd['SAVI']=((1.5 * (dd.nir - dd.red))/(dd.nir + dd.red + 0.5))
    dd['NDWI']=(dd.green - dd.nir)/(dd.green + dd.nir)

    return dd

def custom_function(product, query):

    # Initialise datacube
    dc = datacube.Datacube(app='custom_feature_layers')

    # Load data using query
    ds = dc.load(product,**query)
    
    # band index
    ds['NDVI']=(ds.nir - ds.red)/(ds.nir + ds.red)
    ds['GNDVI']=(ds.nir - ds.green)/(ds.nir + ds.green)
    ds['EVI']=((2.5 * (ds.nir - ds.red))/(ds.nir + 6 * ds.red -7.5 * ds.blue + 1))
    ds['SAVI']=((1.5 * (ds.nir - ds.red))/(ds.nir + ds.red + 0.5))
    ds['NDWI']=(ds.green - ds.nir)/(ds.green + ds.nir)

    return ds

def rice_detect(ds, product, outname_raster, outname_vector, query, field):
    
    ######### Part create Training data ############################################## 
    
    # create polygon 
    vectorize(ds, outname_raster, outname_vector )

    # auto detect the num of CPUs 
    ncpus=round(get_cpu_quota())
       
    # load training data 
    input_data = gpd.read_file(outname_vector) #.shp or geojson
    
    # Extract data using shapfile(polygon)
    column_names, model_input = collect_training_data(gdf=input_data,
                                                      dc_query=query,
                                                      ncpus=ncpus,
                                                      feature_func=custom_loadband,
                                                      field=field,
                                                      zonal_stats='median')
    
    ########### preprocessing training data ##########################################
    print('column_name of training data:',column_names)
  
    # Create a DataFrame from model_input
    df = pd.DataFrame(model_input, columns=column_names)

    # Create a GeoDataFrame from the DataFrame
    gdf = gpd.GeoDataFrame(df)
    
    input_column_names = ['values', 'NDVI','GNDVI', 'EVI', 'SAVI', 'NDWI']
    df = df[input_column_names]

    print(df)
    
    # Define the query conditions
    ndvi_condition = (df['NDVI'].between(0.6, 0.9))
    gndvi_condition = (df['GNDVI'].between(0.28, 0.368))
    savi_condition = (df['SAVI'].between(0.256, 0.648))
    evi_condition = (df['EVI'].between(0.4, 0.8))
    ndwi_condition = (df['NDWI'].between(-0.55,-0.267))
 
    # Modify the input training data for single class labels
    model_input[:, 0] = np.where((ndvi_condition & ~(ndwi_condition|gndvi_condition|savi_condition )) ,1 ,0) 
    #& (savi_condition|gndvi_condition|ndwi_condition)
    
    # convert model_input to dataframe
    df2 = pd.DataFrame(model_input, columns=column_names)
    
    model_input = df2[input_column_names].values
    
    # print(model_input[:, 0])
    # print(input_column_names)
    
    # Split into training and testing data (train80/test20)
    model_train, model_test = model_selection.train_test_split(model_input, 
                                                               stratify=model_input[:, 0], 
                                                               train_size=0.7, 
                                                               random_state=0)
    print('#'*40)
    print("Train shape(70%):", model_train.shape)
    print("Test shape(30%):", model_test.shape)
    print('#'*40)
    
    
    # Convert model_train to a NumPy array
    model_train_array = model_train
    model_test_array = model_test
    ############## Model preparation #################################################

    # Select the variables we want to use to train our model
    model_variables = [col for col in input_column_names if col not in ["values"]]
    
    print('The variables for Training Model :', model_variables)
    
    # Extract relevant indices from the processed shapefile
    model_col_indices = [input_column_names.index(var_name) for var_name in model_variables]
    
    # Initialise model
    model = RandomForestClassifier(n_estimators=500)
    
    print("Initial call RF ML")
    ####################### Train Model/ predict #############################################
    
    # Train model
    model.fit(model_train_array[:, model_col_indices], model_train_array[:, input_column_names.index('values')])
    
    # Predict data
    predictions = model.predict( model_test_array[:, model_col_indices])
    
    print('Training Model!!!')
    
    # show the accuracy of prediction
    acc = accuracy_score(predictions,  model_test_array[:, 0])
    print('The accuracy of model is :',acc)
    
    # This shows the feature importance of the input features for predicting the class labels provided
    plt.bar(x=model_variables, height=model.feature_importances_)
    plt.gca().set_ylabel('Importance', labelpad=10)
    plt.gca().set_xlabel('Variable', labelpad=10);
    
######################## classify #################################################

    shape = input_data
    
    # Compute the bounding box of the union of all geometries in shp
    xmin, ymin, xmax, ymax = shape.unary_union.bounds
    
    # Convert bounding box coordinates to decimal degrees
    gdf_crs_decimal_degrees = shape.to_crs("EPSG:4326")
    xmin, ymin = gdf_crs_decimal_degrees.total_bounds[:2]
    xmax, ymax = gdf_crs_decimal_degrees.total_bounds[2:]

    print('x axis :',xmin, xmax)
    print('y axis :',ymin, ymax)
    
    # Set up the query parameters
    query2 = {
        # 'product': product, 
        'measurements': query['measurements'],
        'time': query['time'],
        'x': (xmin, xmax),
        'y': (ymin, ymax),
        'resolution': query['resolution'], 
        'output_crs': query['output_crs'],
        'group_by': 'solar_day'  }
    
    
    
    # Use custom function to generate input data
    input_data = custom_function(product,query=query2)
    
    
    input_data2 = input_data[['NDVI','GNDVI', 'EVI', 'SAVI', 'NDWI']]
    
    print(input_data2)
    

    # Predict landcover using the trained model
    predicted = predict_xr(model, input_data2, clean=True)
    
    print('predicted!!!')
    
#################################
#     print('*'*60)

#     # Calculate accuracy
#     # acc = accuracy_score(predicted.values, input_data2.values)
#     # print('The accuracy of the classification result is:', acc)
    
#     print('*'*60)
    
#################################### Plot output ############################################## 
    # print(predicted.Predictions)
    
    # Set up plot
    fig, axes = plt.subplots(1, 2, figsize=(15, 6))

    # Plot classified image
    predicted.Predictions.plot(ax=axes[0], 
                   cmap='Reds', 
                   add_labels=False, 
                   add_colorbar=False)

    # Plot true colour image
    (input_data[['red', 'green', 'blue']]
     .squeeze('time')
     .to_array()
     .plot.imshow(ax=axes[1], robust=True, add_labels=False))

    # Remove axis on right plot
    axes[1].get_yaxis().set_visible(False)

    # Add plot titles
    axes[0].set_title('Classified data')
    axes[1].set_title('True colour image');
    
    print('plot image')
    
########################## Export output ###################################

    # Write the predicted data out to a GeoTIFF
    write_cog(predicted.Predictions,
             'predicted.tif',
              overwrite=True)
    print('Export predicted image')
################################################################################    


import datacube
import matplotlib.pyplot as plt
import numpy as np
import scipy
import os
import xarray as xr
import geopandas as gpd
from osgeo import gdal
from osgeo import osr
from osgeo import ogr
from datacube.utils.cog import write_cog
from skimage.segmentation import felzenszwalb
from flusstools.geotools import raster2array, open_raster
from flusstools.geotools import create_raster, get_srs, float2int

# function to convert raster image to polygon 
# by using image segmentation(felzenszwalb)
def vectorize(ds, outname_raster, outname_vector):
    
    # calculate Index to catagorize the value to polygon
    band_diff = 100*((ds.nir - ds.red) /(ds.nir + ds.red))
    
    # select first time step
    time_step = 0
    VI = band_diff.isel(time = time_step)
    
    
    
    # Convert our mean index xarray into a numpy array, we need
    # to be explicit about the datatype to satisfy felzenszwalb
    input_array = VI.values.astype(np.float64)
    
    # Calculate the segments (image segmentation)
    segments = felzenszwalb(input_array, 
                             scale=1, 
                             sigma=0.8, 
                             min_size=50, 
                             channel_axis=-1)
    
    # Calculate the zonal mean index across the segments
    segments_zonal_mean_qs = scipy.ndimage.mean(input=input_array,
                                                labels=segments,
                                                index=segments)
    

    # Convert numpy array to xarray.DataArray
    segments_zonal_mean_qs = xr.DataArray(segments_zonal_mean_qs, 
                                           coords=VI.coords, 
                                           dims=['y', 'x'], 
                                           attrs=VI.attrs)
    
    # Write array to GeoTIFF
    write_cog(geo_im=segments_zonal_mean_qs,
              fname=outname_raster, # .tif
              overwrite=True);
    
    # convert raster to polygon
    raster2polygon(outname_raster , outname_vector) #(.tif,.shp)
    
    #convert CRS
    gdf = gpd.read_file(outname_vector)
    gdf= gdf.to_crs(32647)
    # gdf[gdf.columns[0]] /= 100
    gdf.to_file(outname_vector, driver='ESRI Shapefile', crs=32647, encoding='utf-8')
    
    
    
# func to convert raster image to polygon
def raster2polygon(file_name, out_shp_fn, band_number=1, field_name="values"):
    """
    Convert a raster to polygon
    :param file_name: STR of target file name, including directory; must end on ".tif"
    :param out_shp_fn: STR of a shapefile name (with directory e.g., "C:/temp/poly.shp")
    :param band_number: INT of the raster band number to open (default: 1)
    :param field_name: STR of the field where raster pixel values will be stored (default: "values")
    :return: None
    """
    # if len(out_shp_fn) < 13:
    #     pass
    # else:
    #     raise ValueError('Shapefile name may not have more than 13 characters')
        
    # ensure that the input raster contains integer values only and open the input raster
    file_name = float2int(file_name)
    raster, raster_band = open_raster(file_name, band_number=band_number)

    # create new shapefile with the create_shp function
    new_shp = create_shp(out_shp_fn, layer_name="raster_data", layer_type="polygon")
    dst_layer = new_shp.GetLayer()

    # create new field to define values
    new_field = ogr.FieldDefn(field_name, ogr.OFTInteger)
    dst_layer.CreateField(new_field)

    # Polygonize(band, hMaskBand[optional]=None, destination lyr, field ID, papszOptions=[], callback=None)
    gdal.Polygonize(raster_band, None, dst_layer, 0, [], callback=None)

    # create projection file
    srs = get_srs(raster)
    make_prj(out_shp_fn, int(srs.GetAuthorityCode(None)))
    print("Success: Wrote %s" % str(out_shp_fn)) 
    
 
 # function extension for converting raster image to polygon   
def create_shp(shp_file_dir, overwrite=True, *args, **kwargs):
    """
    :param shp_file_dir: STR of the (relative shapefile directory (ends on ".shp")
    :param overwrite: [optional] BOOL - if True, existing files are overwritten
    :kwarg layer_name: [optional] STR of the layer_name - if None: no layer will be created (max. 13 chars)
    :kwarg layer_type: [optional] STR ("point, "line", or "polygon") of the layer_name - if None: no layer will be created
    :output: returns an ogr shapefile layer
    """
    shp_driver = ogr.GetDriverByName("ESRI Shapefile")

    # check if output file exists if yes delete it
    if os.path.exists(shp_file_dir) and overwrite:
        shp_driver.DeleteDataSource(shp_file_dir)

    # create and return new shapefile object
    new_shp = shp_driver.CreateDataSource(shp_file_dir)

    # create layer if layer_name and layer_type are provided
    if kwargs.get("layer_name") and kwargs.get("layer_type"):
        # create dictionary of ogr.SHP-TYPES
        geometry_dict = {"point": ogr.wkbPoint,
                         "line": ogr.wkbMultiLineString,
                         "polygon": ogr.wkbMultiPolygon}
        # create layer
        try:
            new_shp.CreateLayer(str(kwargs.get("layer_name")),
                                geom_type=geometry_dict[str(kwargs.get("layer_type").lower())])
        except KeyError:
            print("Error: Invalid layer_type provided (must be 'point', 'line', or 'polygon').")
        except TypeError:
            print("Error: layer_name and layer_type must be string.")
        except AttributeError:
            print("Error: Cannot access layer - opened in other program?")
    return new_shp
    
# function to create projection file of a shapfile
def make_prj(shp_file_name, epsg):
    """
    Returns:
        None: Creates a projection file (``.prj``) in the same directory and
        with the same name of ``shp_file_name``.
    """
    shp_dir = shp_file_name.split(shp_file_name.split("/")[-1].split("\\")[-1])
    shp_name = shp_file_name.split(".shp")[0].split("/")[-1].split("\\")[-1]
    with open(shp_dir[0] + shp_name + ".prj", "w+") as prj:
        prj.write(get_wkt(epsg))
        
        
        
def get_wkt(epsg, wkt_format="esriwkt"):
    """Gets WKT-formatted projection information for an epsg code using the ``osr`` library.

    Args:
        epsg (int): epsg Authority code
        wkt_format (str): of wkt format (default is esriwkt for shapefile projections)

    Returns:
        str: WKT (if error: returns default corresponding to ``epsg=4326``).
    """
    default = 'GEOGCS["GCS_WGS_1984",DATUM["D_WGS_1984",SPHEROID["WGS_1984",6378137,298.257223563]],PRIMEM["Greenwich",0],UNIT["Degree",0.017453292519943295],UNIT["Meter",1]]'
    spatial_ref = osr.SpatialReference()
    try:
        spatial_ref.ImportFromEPSG(epsg)
    except TypeError:
        logging.error(
            "epsg must be integer. Returning default WKT(epsg=4326).")
        return default
    except Exception:
        logging.error(
            "epsg number does not exist. Returning default WKT(epsg=4326).")
        return default
    if wkt_format == "esriwkt":
        spatial_ref.MorphToESRI()
    return spatial_ref.ExportToPrettyWkt()

  
    


# -*- coding: utf-8 -*-
# import all function in lib
from dream_river.convertools import *
from dream_river.geobox import *
from dream_river.indices import *
from dream_river.ML import *
from dream_river.plotimg import *


from setuptools import setup, find_packages

DESC ='This is a library that contain rice cultivated area detection tools and other geospatial tools for jupyter environment on https://sphere.gistda.or.th/ in part of Data Cube. Thus, you needs go to sphere.gistda (https://sphere.gistda.or.th/) and register the account to access jupyter lab environment interactively from a browser.'

setup(
    name='dream_river',
    packages=['dream_river'],
    version='0.0.7',
    license='MIT',
    author_email='dreamusaha@gmail.com',
    description= 'This is a library contained rice detection tools and other geospatial tools for jupyter environment on sphere.gistda.or.th in part of Data Cube',
    long_description= DESC,
    author='Pathakorn Usaha',
    url= 'https://github.com/Pathakorn40/rice-detection',
    download_url= 'https://pypi.org/project/dream_river/',
    keywords= ['geography','geospatiol','gis', 'rice detection'],
    install_requires= ['datacube','shapely','xarray','matplotlib','numpy','geopandas','dea_tools','scikit-learn',
                'pydotplus','flusstools','scikit-image','GDAL','scipy','folium','odc.ui'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  
        'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        ],
    )

